/*
	$Id: menu_applet.c,v 1.26 2003/09/06 02:58:46 Treize Exp $
   ROX-Menu
	Simple application menu applet for ROX desktop environment.

   Copyright (C) 2001 Leandro Pereira <leandro@linuxmag.com.br>
	Copyright (C) 2002, 2003 Peter Geer <geerp@cs.sunyit.edu>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <malloc.h>
#include <pwd.h>
#include <signal.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "choices.h"
#include "globals.h"
#include "configwin.h"

#include "Icons/null.xpm"
#include "Icons/group.xpm"

#define SCALED_IMAGE_WIDTH  28
#define SCALED_IMAGE_HEIGHT 28

#define DEFAULT_ICON_WIDTH  60
#define DEFAULT_ICON_HEIGHT 38

#define DEFAULT_GROUPS_ON_BOTTOM 0
#define DEFAULT_STRIP_NUMBER 0

#define TOPLEVEL_NO      0
#define TOPLEVEL_YES     1
#define TOPLEVEL_NO_FILE 2

/* Function Prototypes */
int dir_sort(const void *a, const void *b);
int make_directory_entry(GtkWidget *parent_menu, char *dir, int toplevel);
void create_menu_from_directory(GtkWidget *toplevel_menu);
int build_menu(void);
GtkWidget *new_menu_item(char *dir, char ** from);

/* Global data */
GtkWidget *menu, *window;
gchar* command_strings[1024];
gint cmd_str_len = -1;

static char *get_app_name(char *path)
{
	int last = strlen(path);
	int i = last-1;
	/* Find the end of the directory name. */
	while (path[i] == '/' || path[i] == ' ' || path[i] == '\t') 
		path[i--] = '\0';
	/* Find the last separator. */
	while (path[i] != '/') i--;
	return (char*) path + i + 1;
}

/*
 * Thanks to Carlos Pereira for the example (I think I got this
 * at gtk-app-devel list).
 *
 * Someone there could make this function better? I don't know
 * if it'll work with other GTK themes than the one I'm using
 * due to different font sizes.
 *
 * Fixed: borrowed the code from GnomeICU :P
 */
static void set_position(GtkMenu * _menu, gint * px, gint * py, gpointer data)
{
	GtkWidget *widget;
	gint sw;
	GtkRequisition requisition;

	widget = GTK_WIDGET(_menu);

	gtk_widget_size_request(widget, &requisition);

	gdk_window_get_origin(button->window, px, py);
	*py -= requisition.height;

	sw = gdk_screen_width();

	if (*px + requisition.width > sw)
		*px -= *px + requisition.width - sw;
}

static int run_app(GtkWidget * widget, gpointer data)
{
	char tmp[256];
	pid_t child;

	snprintf(tmp, 256, "%s/AppRun", (char*) data);

	child = fork();
	if (child == -1) return (0);
	else if (child == 0) {
		char *argv[2];
		argv[0] = tmp;
		argv[1] = 0;
		execv(tmp, argv);
		_exit(127);
	}
	return child;
}

GtkWidget *new_menu_item(char *dir, char ** from)
{
	GtkWidget *pixmap;
	/* For the default icon case. */
	GdkPixbuf *gdkpixbuf;
	GtkWidget *menuitem;
	char icon_path[256];
	char *dir_label;
	/* To check if a file is open */
	FILE *app_icon=NULL;

	snprintf(icon_path, 256, "%s/.DirIcon", dir);
	/* Now let's try to open the file.  Do .DirIcon first, and fall 
	   back on deprecated AppIcon.xpm if not readable. */
	app_icon=fopen(icon_path,"r");
	if (app_icon == NULL) {
		snprintf(icon_path, 256, "%s/AppIcon.xpm", dir);
		app_icon=fopen(icon_path, "r");
	}
	
	if (from == NULL) {
		pixmap = create_pixmap(icon_path, NULL, SCALED_IMAGE_WIDTH, 
		                       SCALED_IMAGE_HEIGHT, FALSE);
	} else if (app_icon != NULL) {
		pixmap = create_pixmap(icon_path, NULL, SCALED_IMAGE_WIDTH, 
		                       SCALED_IMAGE_HEIGHT, FALSE);
	} else {
		gdkpixbuf = gdk_pixbuf_new_from_xpm_data((const char **)from);
		pixmap = gtk_image_new_from_pixbuf(gdkpixbuf);
		if (GDK_IS_PIXBUF(gdkpixbuf)) gdk_pixbuf_unref(gdkpixbuf);
	}
	if (app_icon != NULL) fclose(app_icon);
	
	gtk_widget_show(pixmap);

	dir_label = get_app_name(dir);
	if (strip_number && isdigit(dir_label[0])>0 ) {
		dir_label = strchr(dir_label, '_') + 1;
	}

	menuitem = gtk_image_menu_item_new_with_label(dir_label);
	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem), pixmap);

	gtk_widget_show(menuitem);

	return menuitem;

}

int dir_sort(const void *a, const void *b) {
	gint ret;
	gchar *s1=NULL, *s2=NULL;

	s1 = g_utf8_casefold((*(const struct dirent **) a)->d_name, -1);
	s2 = g_utf8_casefold((*(const struct dirent **) b)->d_name, -1);
	
	ret = g_utf8_collate(s1, s2);
	
	g_free(s1);
	g_free(s2);
	
	return ret;
/*	return  g_ascii_strcasecmp((*(const struct dirent **) a)->d_name, 
	                           (*(const struct dirent **) b)->d_name);*/
}

int make_directory_entry(GtkWidget *parent_menu, char *dir, int toplevel) {

	GtkWidget *menuitem, *submenu, *submenuitems;
	char tmp[256], tmpdir[256], buf[256];
	gchar *command_text;
	int i=0, j=0;
	int num_entries;
	int has_dirs = 0, has_apps = 0;
	int ret = 0, chk = 0;
	struct stat info;
	DIR *dp;
	struct dirent **entries;
	struct dirent *temp;
	
	snprintf(buf, 256, "%s", dir);
	if (!(dp = opendir(dir))) {
		fprintf(stderr, "*** Can't opendir(\"%s\").\n", dir);
		return 0;
	}
	chk = closedir(dp);
	if (chk != 0) fprintf(stderr, "Error closing directory. %d\n", chk); 
	
	if (toplevel == TOPLEVEL_NO_FILE) {
		submenu = parent_menu;
	} else {
		submenu = gtk_menu_new();
	}
	gtk_widget_show(submenu);
	
	num_entries = scandir(dir, &entries, 0, dir_sort);
	
	/* If we're putting groups on the bottom, we have to reverse
	   the list so that the prepends come out in the right order.*/
	if (groups_on_bottom) {
		i = 0;
		j = num_entries - 1;
		while (i<j) {
			temp = entries[i];
			entries[i] = entries[j];
			entries[j] = temp;
			i++;
			j--;
		}
	}
	
	i=num_entries-1;
	
	while (i>=0) {
		snprintf(tmp, 256, "%s/%s/AppRun", buf, entries[i]->d_name);
		snprintf(tmpdir, 256, "%s/%s", buf, entries[i]->d_name);
		if (!strcmp(entries[i]->d_name, "..") || 
		    !strcmp(entries[i]->d_name, ".") ||
			 (entries[i]->d_name[0] == '.') ) {
			i--;
			continue;
		}
		/* If the directory is an AppDir, don't add a submenu. */
		if (!stat(tmp, &info) ) {
			has_apps = 1;
			i--;
			continue;
		/* If it's a directory, but not an AppDir, add a submenu. */
		} else {

			dp=opendir(tmpdir);
			if (dp) {
				/* Recursively add submenus.  This allows us to have an
				arbitrary number of nested submenus, but risks infinite
				looping if you're dumb enough to link back to a point 
				higher in the directory tree. */
				ret = make_directory_entry(submenu, tmpdir, 0);
				has_dirs = ret;
			chk = closedir(dp);
			if (chk != 0) fprintf(stderr, "Error closing directory%d", chk);
			}
		}
		i--;
	}

	if (has_apps && has_dirs) {
		submenuitems = gtk_separator_menu_item_new();
		gtk_widget_show(submenuitems);
		gtk_widget_set_sensitive(GTK_WIDGET(submenuitems), FALSE);
		if (groups_on_bottom) {
			gtk_menu_prepend(GTK_MENU(submenu), submenuitems);
		} else {	
			gtk_menu_append(GTK_MENU(submenu), submenuitems);
		}
	}

	i=0;
	
	while (i < num_entries) {
		snprintf(tmp, 256, "%s/%s/AppRun", buf, entries[i]->d_name);
		if (!stat(tmp, &info)) {
			ret = 1;  /* Return TRUE if there are any AppDirs. */
			snprintf(tmp, 256, "%s/%s", buf, entries[i]->d_name);
			submenuitems = new_menu_item(tmp, null_xpm);
			command_text =  g_strdup(tmp);
			gtk_signal_connect(GTK_OBJECT(submenuitems), "activate", (GtkSignalFunc) run_app, command_text);
			command_strings[++cmd_str_len] = command_text;
			if (groups_on_bottom) {
				gtk_menu_prepend(GTK_MENU(submenu), submenuitems);
			} else {
				gtk_menu_append(GTK_MENU(submenu), submenuitems);
			}
		}
		i++;
	}

	free(entries);

	if (toplevel == TOPLEVEL_NO_FILE) return ret;
	
	/* If there were any AppDirs, add the submenu groups.
	   Otherwise, destroy them. */
	if (ret) {
		menuitem = new_menu_item(dir, group_xpm);
		if (groups_on_bottom || toplevel == TOPLEVEL_YES) {
			gtk_menu_append(GTK_MENU(parent_menu), menuitem);
		} else {
			gtk_menu_prepend(GTK_MENU(parent_menu), menuitem);
		}
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), submenu);
	} else {
		gtk_widget_destroy(GTK_WIDGET(submenu) );
	}
	return ret;
}

void create_menu_from_directory(GtkWidget *toplevel_menu) {

	guchar *choicesdir;
	gchar *menu_dir_path, *dest_dir, *home_apps;
	int ret;

	choicesdir = choices_find_path_load(NULL, app_name);
	if (choicesdir == NULL) 
		choicesdir = choices_find_path_save(NULL, app_name, FALSE);
	menu_dir_path = g_strconcat(choicesdir, "Menu_Root", NULL);

	if (g_file_test(menu_dir_path, G_FILE_TEST_IS_DIR)) {
		make_directory_entry(toplevel_menu, menu_dir_path, TOPLEVEL_NO_FILE);
	} else {
		/* The directory does not exist, so try to create it with some
		   reasonable default links. */
		ret = mkdir(choicesdir, 0777);
		ret = mkdir(menu_dir_path, 0777);
		if (ret != 0) {
			fprintf(stderr, "Fatal error: Unable to create configuration directory %s.\n",
			        menu_dir_path);
			exit(1);
		}
		
		dest_dir = g_strconcat(menu_dir_path, "/SystemApps", NULL);
		ret = symlink("/usr/apps", dest_dir);
		g_free(dest_dir);
		dest_dir = g_strconcat(menu_dir_path, "/LocalApps", NULL);
		ret = symlink("/usr/local/apps", dest_dir);
		g_free(dest_dir);
		dest_dir = getenv("HOME");
		home_apps = g_strconcat(dest_dir, "/Apps", NULL);
		g_free(dest_dir);
		dest_dir = g_strconcat(menu_dir_path, "/UserApps", NULL);
		ret = symlink(home_apps, dest_dir);
		g_free(dest_dir);
		g_free(home_apps);
		make_directory_entry(toplevel_menu, menu_dir_path, TOPLEVEL_NO_FILE);

	}
	g_free(menu_dir_path);
	g_free(choicesdir);
}

int build_menu(void) {

	GtkWidget *menuitem;
	FILE *apps;
	char buffer[1024], *strip, *end, *val[3];
	guchar *user_choices;
	gchar* command_text;

	while (cmd_str_len >= 0) g_free(command_strings[cmd_str_len--]);

	/* I don't think this is working like it's supposed to... -P.G.*/
	if (GTK_IS_WIDGET(menu)) gtk_widget_destroy(GTK_WIDGET(menu));

	menu = gtk_menu_new();

	user_choices = choices_find_path_load("Apps", app_name);

	if (user_choices == NULL) {
		create_menu_from_directory(menu);
		return (1);
		/*
		create_choices_file();
		user_choices = choices_find_path_load("Apps", app_name);
		*/
	}

	apps = fopen(user_choices, "r");

	while (fgets((char *) buffer, 1024, apps)) {
		
		if ((strip = strchr((char *) buffer, '\n'))) *strip = 0;
		/* Make everything after a '#' into a comment. */
		end = strchr(buffer, '#');
		if (end) *end = 0;
		if (!strlen(buffer)) continue;

		val[0] = strtok(buffer, "=");
		val[1] = strtok(NULL, "=");

		if (!strncmp(buffer, "-", 1)) {	/* Separator */
			menuitem = gtk_separator_menu_item_new();

			gtk_widget_show(menuitem);
			gtk_widget_set_sensitive(GTK_WIDGET(menuitem), FALSE);

			gtk_menu_append(GTK_MENU(menu), menuitem);
 
		} else if (!strncmp(buffer, "d", 1)) {		/* Group */
		
			make_directory_entry(menu, val[1], 1);
		
		} else if(!strncmp(buffer, "l", 1)) {  /* Label */
			GtkStyle *style;
			GtkWidget *label;

			label=gtk_label_new(val[1]);
			gtk_widget_show(GTK_WIDGET(label));
	    
			menuitem=gtk_menu_item_new();
			gtk_widget_show(GTK_WIDGET(menuitem));

			/* Does this actually _do_ anything? */
			style=gtk_style_copy(gtk_widget_get_style(label));
			gtk_widget_set_style(GTK_WIDGET(label), style);

			gtk_container_add(GTK_CONTAINER(menuitem), label);
			gtk_menu_append(GTK_MENU(menu), menuitem);
			
		} else {		/* Normal menu item */

			menuitem = new_menu_item(val[1], NULL);
			command_text = g_strdup(val[1]);
			gtk_signal_connect(GTK_OBJECT(menuitem), "activate", (GtkSignalFunc) run_app, command_text);
			command_strings[++cmd_str_len] = command_text;
			gtk_menu_append(GTK_MENU(menu), menuitem);

		}
   }
	
   g_free(user_choices);
   return (1);
}

/* This is called as a signal handler */
static void child_died(int signum)
{
	pid_t	child;
	int	status;
	/* Find out which children exited and allow them to die */
	while (1)
	{
		child = waitpid(-1, &status, WNOHANG);
		if (child == 0 || child == -1)
			break;
	}
}

/* 
	Some signal handlers for destroying the window.  I found it annoying
	that clicking the WM's 'X' left the program running.	-P.G.
*/
static gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data) {
	return(FALSE);
}

static void quit_app(void) {
	write_config();
	gtk_main_quit();
}

static void destroy(GtkWidget *widget, gpointer data) { 
	quit_app(); 
}

static void refresh_menu(void) {
	build_menu();	
}

static void show_choices(void) {
	char cmd[256];
	guchar *choices_path = choices_find_path_load(NULL, app_name);
	snprintf(cmd, 256, "rox %s", choices_path);
	system(cmd);
	g_free(choices_path);
}

/* Menu stuff borrowed from DemoApplet by Stephen Watson. */
static GtkItemFactoryEntry menu_items[] = {
  { ("/Refresh"), NULL, refresh_menu, 0, NULL },
  { ("/Options..."), NULL, config_win_show, 0, NULL },
  { ("/Choices dir..."), NULL, show_choices, 0, NULL },
  { ("/Quit"), NULL, quit_app, 0, NULL }
};

static GtkWidget *menu_create_menu(GtkWidget *window)
{
  GtkWidget *context_menu;
  GtkItemFactory	*item_factory;
  GtkAccelGroup	*accel_group;
  gint  n_items = sizeof(menu_items) / sizeof(*menu_items);

  accel_group = gtk_accel_group_new();

  item_factory = gtk_item_factory_new(GTK_TYPE_MENU, "<system>", 
				      accel_group);

  gtk_item_factory_create_items(item_factory, n_items, menu_items, NULL);

  /* Attach the new accelerator group to the window. */
  gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);

  context_menu = gtk_item_factory_get_widget(item_factory, "<system>");

  return context_menu;
}

static gint button_press(GtkWidget *btn, GdkEventButton *bev, gpointer win) {

	static GtkWidget *sys_menu;
	
	if (bev->type == GDK_BUTTON_PRESS && bev->button == 3) {

		if (sys_menu == NULL) sys_menu=menu_create_menu(GTK_WIDGET(win));
			gtk_menu_popup(GTK_MENU(sys_menu), NULL, NULL, NULL, NULL, 
			               bev->button, bev->time);
		return TRUE;
	
	} else if (bev->type == GDK_BUTTON_PRESS && bev->button == 1) {

		gtk_menu_popup(GTK_MENU(menu), NULL, NULL, (void*)set_position, 
		               NULL, bev->button, bev->time);
		gtk_widget_set_sensitive(GTK_WIDGET(button), TRUE);
		return TRUE;
	
	}
	return FALSE;
}

int main(int argc, char *argv[], char * envp[])
{
	GtkTooltips *tip = NULL;
	struct sigaction	act;

	 /* Initialize global data with default values. */
	icon_width = DEFAULT_ICON_WIDTH;   /* This size works pretty well  */
	icon_height = DEFAULT_ICON_HEIGHT; /* with the new ROX eagle pixmap*/
	groups_on_bottom = DEFAULT_GROUPS_ON_BOTTOM;
	strip_number = DEFAULT_STRIP_NUMBER;
	menu_file_on = g_strdup(" ");
	menu_file_off = g_strdup(" ");
	app_name = getenv("APP_NAME");

	gtk_init(&argc, &argv);
	gdk_rgb_init();
	gtk_widget_push_visual(gdk_rgb_get_visual());
	gtk_widget_push_colormap(gdk_rgb_get_cmap());
    
	if (argc > 1) {
		window = gtk_plug_new(atol(argv[1]));
	} else {
		window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	}
	
	gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(delete_event), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(destroy), NULL);

	/* I have no idea what this does.... -- P.G. */
	/* Let child processes die */
	act.sa_handler = child_died;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_NOCLDSTOP;
	sigaction(SIGCHLD, &act, NULL);
	choices_init();
	read_config();
	config_win_init();

	/*gtk_window_set_default_size(GTK_WINDOW(window), APPLET_SIZE, APPLET_SIZE);*/
	gtk_window_set_title(GTK_WINDOW(window), app_name);

	button = gtk_button_new();
	gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
	GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(button), GTK_CAN_FOCUS);
	gtk_container_set_border_width(GTK_CONTAINER(button), 0);

	tip = gtk_tooltips_new();
	gtk_tooltips_set_tip(GTK_TOOLTIPS(tip), button,
	                     "Click to start ROX applications.",
								"Displays a menu of applications.  Right-click to configure, refresh, or quit the menu.");

	change_menu_pixmap(LEAVE);

	gtk_widget_show(button);
	gtk_container_add(GTK_CONTAINER(window), button);

	gtk_widget_add_events(GTK_WIDGET(button), GDK_BUTTON_PRESS_MASK);
	gtk_signal_connect(GTK_OBJECT(button), "button_press_event", 
	                   GTK_SIGNAL_FUNC(button_press), window);
	gtk_signal_connect_object(GTK_OBJECT(button), "enter", 
	                          GTK_SIGNAL_FUNC(change_menu_pixmap),
	                          GINT_TO_POINTER(ENTER));
	gtk_signal_connect_object(GTK_OBJECT(button), "leave", 
	                          GTK_SIGNAL_FUNC(change_menu_pixmap), 
	                          GINT_TO_POINTER(LEAVE));

	build_menu();
	gtk_widget_show(window);

	gtk_main();

	return (0);
}
